/*
 * Scheme_3 compiler runtime
 * By Nils M Holm, 2016
 *
 * Provided under the Creative Common Zero (CC0) license
 * (https://creativecommons.org/publicdomain/zero/1.0/)
 *
 * From the book COMPILING LAMBDA CALCULUS, http://t3x.org/clc/
 */

#include "s9core.h"
#include "s9import.h"

#define	SEG	1000

#define box(x)	cons((x),NIL)
#define unbox	car

cell	E, L, S, T;
int	K, P, Z;
char	*I;

cell	*R[] = { &E, &L, &S, &T, NULL };

void err(char *s) {
	fprintf(stderr, "LC8 error: %s\n", s);
	exit(1);
}

void err1(void) {
	err("wrong number of arguments");
}

void init(void) {
	s9_init(R);
	P = 0;
	Z = SEG;
	S = make_vector(Z);
}

cell readexp(void) {
	#define B 256
	char	b[B];
	int	i;
	cell	n, a, p;

	while (isspace(*I))
		I++;
	if ('(' == *I) {
		I++;
		a = NIL;
		n = readexp();
		while (n != VOID) {
			if (NIL == a) {
				save(a = cons(NIL, NIL));
			}
			else {
				p = cons(NIL, NIL);
				cdr(a) = p;
				a = cdr(a);
			}
			car(a) = n;
			n = readexp();
		}
		if (NIL == a)
			return NIL;
		return unsave(1);
	}
	else if (')' == *I) {
		I++;
		return VOID;
	}
	else {
		for (i = 0; !isspace(*I) && ')' != *I; i++) {
			b[i] = *I++;
			if (i >= B) err("symbol too long");
		}
		b[i] = 0;
		return symbol_ref(b);
	}
}

void px(cell x) {
	if (function_p(x)) {
		printf("#<function %d>", (int) cadr(x));
	}
	else if (symbol_p(x)) {
		printf("%s", symbol_name(x));
	}
	else if (NIL == x) {
		printf("()");
	}
	else {
		putchar('(');
		while (!atom_p(x)) {
			px(car(x));
			x = cdr(x);
			if (pair_p(x))
				putchar(' ');
		}
		if (x != NIL) {
			printf(" . ");
			px(x);
		}
		putchar(')');
	}
}

void printexp(cell x) {
	px(x);
	putchar('\n');
}

cell mklitpool(char *s) {
	cell	n, v, *vv;
	int	i;

	I = s;
	n = readexp();
	save(n);
	v = make_vector(length(n));
	unsave(1);
	vv = vector(v);
	for (i = 0; n != NIL; n = cdr(n))
		vv[i++] = car(n);
	return v;
}

void push(cell x) {
	cell	v;

	if (P >= Z) {
		T = x;
		v = make_vector(Z+SEG);
		memcpy(vector(v), vector(S), Z*sizeof(cell));
		Z += SEG;
		S = v;
	}
	vector(S)[P++] = x;
}

cell pop(void) {
	if (P < 0)
		err("stack underflow");
	return vector(S)[--P];
}

void pushval(int n) {
	push(new_atom(n, NIL));
}

int popval(void) {
	return car(pop());
}

void pushfun(int k, int n, int z) {
	cell	fn, ne;

	ne = make_vector(z);
	memcpy(vector(ne), vector(E), n*sizeof(cell));
	fn = new_atom(k, ne);
	push(new_atom(T_FUNCTION, fn));
}

void collect_args(int na) {
	cell	k, e, va;
	int	n, i;

	k = vector(S)[P-1];
	e = vector(S)[P-2];
	n = car(vector(S)[P-3]);
	if (n < na-1)
		err1();
	save(va = NIL);
	for (i = 0; n-i >= na; i++) {
		va = cons(vector(S)[P-4-i], va);
		car(Stack) = va;
	}
	va = unsave(1);
	P = P-i+1;
	vector(S)[P-1] = k;
	vector(S)[P-2] = e;
	vector(S)[P-4] = va; /* save va first! */
	vector(S)[P-3] = box(na);
}

void save_vars(int n, int k) {
	cell	v;
	int	i;

	v = make_vector(k-n);
	for (i=n; i<k; i++)
		vector(v)[i-n] = vector(E)[i];
	push(v);
}

void restore_vars(cell v) {
	int	kv, ke, i;

	kv = vector_len(v);
	ke = vector_len(E);
	for (i=0; i<kv; i++)
		vector(E)[ke-kv+i] = vector(v)[i];
}

int apply(int k, int tco) {
	int	m, n;
	cell	e, kk;

	T = pop();
	if (!function_p(T))
		err("application of non-function");
	if (tco) {
		m = unbox(vector(S)[P-1])+1;
		restore_vars(vector(S)[P-3]);
		kk = vector(S)[P-m-2];
		e = vector(S)[P-m-3];
		n = unbox(vector(S)[P-m-4])+1;
		memcpy(&vector(S)[P-m-n-3],
			&vector(S)[P-m],
			m*sizeof(cell));
		P -= n+1;
		vector(S)[P-1] = kk;
		vector(S)[P-2] = e;
		E = cddr(T);
		return cadr(T);
	}
	push(E);
	pushval(k);
	E = cddr(T);
	return cadr(T);
}

int ret(void) {
	int	k, n;

	T = pop();
	restore_vars(pop());
	k = popval();
	E = pop();
	n = popval();
	P -= n;
	push(T);
	return k;
}
